<?PHP
#
#   FILE:  StartUp.php
#
#   USAGE:  This file initializes the CWIS operating environment, and may be
#       included in non-CWIS files that use the CWIS user login mechanism or
#       other portions of the CWIS application framework.
#
#   OPTIONS
#   These global variables may be set before including the file:
#       StartUpOpt_CLEAR_AF_CACHES - When set to TRUE, all ApplicationFramework
#           caches will be cleared right after $GLOBALS["AF"] is loaded.
#       StartUpOpt_DO_NOT_LOAD_PLUGINS - When set to TRUE, plugins will not
#           be loaded at all.  (Though PluginManager will still be created.)
#       StartUpOpt_FORCE_PLUGIN_CONFIG_LOAD - When set to TRUE, all plugins will
#           be instructed to set up their configuration options, even if the
#           plugin is not currently enabled.
#       StartUpOpt_USE_AXIS_USER - When set to TRUE, $GLOBALS["G_User"] will
#           be a ScoutLib User object rather than a CWUser.
#
#   Part of the Collection Workflow Integration System (CWIS)
#   Copyright 2011-2017 Edward Almasy and Internet Scout Research Group
#   http://scout.wisc.edu/cwis
#

# save our current directory then change to CWIS base directory
$StartingDir = getcwd();
chdir(dirname(__FILE__)."/..");

# turn error reporting up to max if running local or at Scout
if (isset($_SERVER["HTTP_HOST"])
        && (($_SERVER["HTTP_HOST"] == "localhost")
                || ($_SERVER["HTTP_HOST"] == "test.scout.wisc.edu")))
{
    error_reporting(E_ALL|E_STRICT);
    # (set time zone to prevent PHP local time zone warning from E_STRICT)
    date_default_timezone_set("America/Chicago");
}

# if configuration file available in expected location
if (file_exists("local/config.php"))
{
    # load configuration file
    require_once("local/config.php");
}
else
{
    # load configuration file from legacy location
    require_once("config.php");

}

# set configuration from legacy variables if necessary
if (!isset($GLOBALS["G_Config"]))
{
    $GLOBALS["G_Config"]["Database"] = array(
            "UserName" => $GLOBALS["SPT_DBUserName"],
            "Password" => $GLOBALS["SPT_DBPassword"],
            "Host" => $GLOBALS["SPT_DBHost"],
            "DatabaseName" => $GLOBALS["SPT_DBName"],
            );
    $GLOBALS["G_Config"]["UserInterface"] = array(
            "DefaultUI" => $GLOBALS["SPT_DefaultUI"],
            );
}

# set up database access
require_once("lib/ScoutLib/Database.php");
Database::SetGlobalServerInfo($GLOBALS["G_Config"]["Database"]["UserName"],
        $GLOBALS["G_Config"]["Database"]["Password"],
        $GLOBALS["G_Config"]["Database"]["Host"]);
Database::SetGlobalDatabaseName($GLOBALS["G_Config"]["Database"]["DatabaseName"]);

# turn on database error display if PHP error display is enabled
if (get_cfg_var("display_errors"))
{
    Database::DisplayQueryErrors(TRUE);
}

# set default storage engine for database server to MyISAM so any
#       new tables support full text searching
$DB = new Database();
$DB->SetDefaultStorageEngine("MyISAM");

# increase session data timeout to eight hours
require_once("lib/ScoutLib/ApplicationFramework.php");
ApplicationFramework::SessionLifetime(8 * 60 * 60);

# initialize application framework
ApplicationFramework::AddObjectDirectory("lib/ScoutLib");
ApplicationFramework::AddObjectDirectory("lib/Parsedown");
ApplicationFramework::AddObjectDirectory("lib/PHPMailer", array(),
        function ($ClassName)
        {
            return "class.".strtolower($ClassName);
        });
$GLOBALS["AF"] = new ApplicationFramework();
if (array_key_exists("StartUpOpt_CLEAR_AF_CACHES", $GLOBALS)
        && $GLOBALS["StartUpOpt_CLEAR_AF_CACHES"])
{
    $GLOBALS["AF"]->ClearTemplateLocationCache();
    $GLOBALS["AF"]->ClearObjectLocationCache();
    $GLOBALS["AF"]->ClearPageCache();
}
$GLOBALS["AF"]->LogFile("local/logs/cwis.log");
require_once("include/Events.php");
$GLOBALS["AF"]->AddEnvInclude("include/Compat.php");
$GLOBALS["AF"]->AddIncludeDirectories(array(
        "lib/CKEditor",
        "lib/D3",
        "lib/C3",
        "lib/jsbn",
        "lib/jquery",
        "lib/jquery-ui",
        ), TRUE);
$GLOBALS["AF"]->DoNotUrlFingerprint("%lib/CKEditor%");

# set up environment
require_once("include/Compat.php");
require_once("include/StdLib.php");
require_once("include/EnvSetup.php");
$GLOBALS["AF"]->HtmlCharset($GLOBALS["G_SysConfig"]->DefaultCharacterSet());
SearchParameterSet::SetCanonicalFieldFunction(
        array("MetadataSchema", "GetCanonicalFieldIdentifier"));
SearchParameterSet::SetPrintableFieldFunction(
        array("MetadataSchema", "GetPrintableFieldName"));
SearchParameterSet::SetLegacyUrlTranslationFunction(
        array("MetadataSchema", "TranslateLegacySearchValues"));
SearchParameterSet::SetTextDescriptionFilterFunction(
        array("SPTSearchEngine", "FilterTextDisplay"));

# set UI for logged-in user (or default UI if not logged in)
$ActiveUI = $GLOBALS["G_SysConfig"]->DefaultActiveUI();
if ($GLOBALS["G_User"]->IsLoggedIn()
        && ($GLOBALS["G_SysConfig"]->AllowMultipleUIsEnabled()
                || $GLOBALS["G_User"]->HasPriv(PRIV_SYSADMIN)))
{
    $UserUI = $GLOBALS["G_User"]->Get("ActiveUI");
    if (!is_null($UserUI))
    {
        $ActiveUI = $UserUI;
    }
}
$AF->ActiveUserInterface($ActiveUI);

# load plugins
PluginManager::SetConfigValueLoader(array("FormUI", "LoadValue"));
$GLOBALS["G_PluginManager"] = new PluginManager($GLOBALS["AF"],
        array("plugins", "local/plugins"),
        (array_key_exists("StartUpOpt_FORCE_PLUGIN_CONFIG_LOAD", $GLOBALS)
                && $GLOBALS["StartUpOpt_FORCE_PLUGIN_CONFIG_LOAD"]));
if (!array_key_exists("StartUpOpt_DO_NOT_LOAD_PLUGINS", $GLOBALS)
        || !$GLOBALS["StartUpOpt_DO_NOT_LOAD_PLUGINS"])
{
    $GLOBALS["G_PluginManager"]->LoadPlugins();
}

# reload environmental settings so that any system configuration changes made
#       by plugins will be reflected
ApplicationFramework::PreferHttpHost($GLOBALS["G_SysConfig"]->PreferHttpHost());
if ($GLOBALS["G_SysConfig"]->RootUrlOverride() !== NULL)
{
    ApplicationFramework::RootUrlOverride($GLOBALS["G_SysConfig"]->RootUrlOverride());
}
Email::DeliverySettings($GLOBALS["G_SysConfig"]->Value("EmailDeliverySettings"));
Email::DefaultFrom(trim($GLOBALS["G_SysConfig"]->PortalName())
        ." <".trim($GLOBALS["G_SysConfig"]->AdminEmail()).">");
InitializeEmailLineEnding($GLOBALS["G_SysConfig"]);
SPTSearchEngine::SetUpdatePriority(
        $GLOBALS["G_SysConfig"]->SearchEngineUpdatePriority());
SPTRecommender::SetUpdatePriority(
        $GLOBALS["G_SysConfig"]->RecommenderEngineUpdatePriority());
SPTSearchEngine::SetNumResourcesForFacets(
        $GLOBALS["G_SysConfig"]->NumResourcesForSearchFacets());

# set ownership of metadata fields by plugins
MetadataSchema::SetOwnerListRetrievalFunction(
        array($GLOBALS["G_PluginManager"], "GetActivePluginList"));
MetadataSchema::NormalizeOwnedFields();

# return to original directory
chdir($StartingDir);
unset($StartingDir);
