<?PHP
#
#   FILE:  FormUI_Base.php
#
#   Part of the Collection Workflow Integration System (CWIS)
#   Copyright 2016 Edward Almasy and Internet Scout Research Group
#   http://scout.wisc.edu/cwis/
#

/**
* Base class (covering non-presentation elements) supplying a standard user
* interface for presenting and working with HTML forms.
*/
abstract class FormUI_Base
{
    # ---- PUBLIC INTERFACE --------------------------------------------------

    /** Supported field types. */
    const FTYPE_FILE = "File";
    const FTYPE_FLAG = "Flag";
    const FTYPE_IMAGE = "Image";
    const FTYPE_METADATAFIELD = "MetadataField";
    const FTYPE_NUMBER = "Number";
    const FTYPE_OPTION = "Option";
    const FTYPE_PARAGRAPH = "Paragraph";
    const FTYPE_PASSWORD = "Password";
    const FTYPE_PRIVILEGES = "Privileges";
    const FTYPE_SEARCHPARAMS = "Search Parameters";
    const FTYPE_TEXT = "Text";
    const FTYPE_URL = "URL";
    const FTYPE_USER = "User";
    const FTYPE_QUICKSEARCH = "Quick Search";
    /** Supported field pseudo-types. */
    const FTYPE_HEADING = "Heading";

    /**
    * Class constructor.
    * @param array $FieldParams Associative array of associative arrays of
    *       form field parameters, with field names for the top index.
    * @param array $FieldValues Associative array of current values for
    *       form fields, with field names for the index.  (OPTIONAL, as values
    *       may also be supplied via $FieldParams.)
    * @param string $UniqueKey Unique string to include in form field names
    *       to distinguish them from other fields in the form.  (OPTIONAL)
    */
    public function __construct(
            $FieldParams, $FieldValues = array(), $UniqueKey = NULL)
    {
        # make sure parameters are legal and complete
        $BooleanParams = array(
                "AllowMultiple",
                "ReadOnly",
                "Required",
                "UseWYSIWYG",
                );
        foreach ($FieldParams as $FieldName => $Params)
        {
            if (!isset($Params["Type"]))
            {
                $ErrMsgs[] = "Type missing for form field ".$FieldName.".";
            }
            elseif (($Params["Type"] == self::FTYPE_OPTION)
                    && !isset($Params["Options"]))
            {
                $ErrMsgs[] = "Option values missing for form field ".$FieldName.".";
            }
            elseif ($Params["Type"] == self::FTYPE_QUICKSEARCH)
            {
                if (!isset($Params["Field"]))
                {
                    $ErrMsgs[]= "Field to search missing for quicksearch form field "
                        .$FieldName.".";
                }
                elseif (!MetadataSchema::FieldExistsInAnySchema($Params["Field"]))
                {
                    $ErrMsgs[]= "Specified search field for quicksearch form field "
                        .$FieldName." does not exist.";
                }
                else
                {
                    $FieldParams[$FieldName]["Field"] =
                        MetadataSchema::GetCanonicalFieldIdentifier($Params["Field"]);
                    $MField = new MetadataField(
                        $FieldParams[$FieldName]["Field"]);

                    $AllowedMFieldTypes = [
                        MetadataSchema::MDFTYPE_TREE,
                        MetadataSchema::MDFTYPE_CONTROLLEDNAME,
                        MetadataSchema::MDFTYPE_USER,
                        MetadataSchema::MDFTYPE_REFERENCE ];
                    if (!in_array($MField->Type(), $AllowedMFieldTypes))
                    {
                        $ErrMsgs[]= "Specified search field for quicksearch form field "
                        .$FieldName." is not a type that supports quicksearches.";
                    }
                }
            }

            if (!isset($Params["Label"]))
            {
                $ErrMsgs[] = "Label missing for form field ".$FieldName.".";
            }
            if (isset($Params["ValidateFunction"])
                    && !is_callable($Params["ValidateFunction"]))
            {
                $ErrMsgs[] = "Uncallable validation function for form field "
                        .$FieldName.".";
            }
            if (isset($Params["InsertIntoField"])
                    && !isset($FieldParams[$Params["InsertIntoField"]]))
            {
                $ErrMsgs[] = "Unknown insertion field (".$Params["InsertIntoField"]
                        .") found for form field ".$FieldName.".";
            }
            foreach ($BooleanParams as $ParamName)
            {
                if (!isset($Params[$ParamName]))
                {
                    $FieldParams[$FieldName][$ParamName] = FALSE;
                }
            }
        }
        if (isset($ErrMsgs))
        {
            $ErrMsgString = implode("  ", $ErrMsgs);
            throw new InvalidArgumentException($ErrMsgString);
        }

        # save form parameters and values
        $this->FieldParams = $FieldParams;
        $this->FieldValues = $FieldValues;
        $this->UniqueKey = $UniqueKey;
    }

    /**
    * Display HTML table with settings parameters.
    * @param string $TableId CSS ID for table element.  (OPTIONAL)
    * @param string $TableStyle CSS styles for table element.  (OPTIONAL)
    */
    abstract public function DisplayFormTable($TableId = NULL, $TableStyle = NULL);

    /**
    * Log error message for later display.
    * @param string $Msg Error message.
    * @param string $Field Field associated with error.  (OPTIONAL, defaults
    *       to no field association)
    */
    public static function LogError($Msg, $Field = NULL)
    {
        self::$ErrorMessages[$Field][] = $Msg;
    }

    /**
    * Get logged errors.
    * @return array Logged errors, with associated fields for the index (NULL
    *       for errors with no association) and an array of error messages for
    *       each value.
    */
    public static function GetLoggedErrors()
    {
        return self::$ErrorMessages;
    }

    /**
    * Report whether errors have been logged.
    * @param string $Field Field to check -- specify NULL to check for any
    *       errors with no field associated.  (OPTIONAL)
    * @return bool TRUE if errors have been logged, otherwise FALSE.
    */
    public static function ErrorsLogged($Field = FALSE)
    {
        if ($Field === FALSE)
        {
            return count(self::$ErrorMessages) ? TRUE : FALSE;
        }
        else
        {
            return isset(self::$ErrorMessages[$Field]) ? TRUE : FALSE;
        }
    }

    /**
    * Clear logged errors.
    * @param string $Field Clear only errors for specified field.  (OPTIONAL)
    */
    public static function ClearLoggedErrors($Field = FALSE)
    {
        if ($Field === FALSE)
        {
            self::$ErrorMessages = array();
        }
        else
        {
            unset(self::$ErrorMessages[$Field]);
        }
    }

    /**
    * Validate field values on submitted form.  Validation functions (specified
    * via the "ValidateFunction" parameter) should take a field name and value
    * as parameters, and return NULL if the field validates successfully, or
    * an error message if it does not.
    * @return int Number of fields with invalid values found.
    */
    public function ValidateFieldInput()
    {
        # retrieve field values
        $Values = $this->GetNewValuesFromForm();

        # for each field
        $ErrorsFound = 0;
        foreach ($this->FieldParams as $Name => $Params)
        {
            # skip header fields
            if ($Params["Type"] == self::FTYPE_HEADING) {  continue;  }

            # determine if field has a value set
            switch ($Params["Type"])
            {
                case self::FTYPE_SEARCHPARAMS:
                    $IsEmpty = !$Values[$Name]->ParameterCount();
                    break;

                case self::FTYPE_PRIVILEGES:
                    $IsEmpty = !$Values[$Name]->ComparisonCount();
                    break;

                default:
                    if (is_array($Values[$Name]))
                    {
                        $IsEmpty = !count($Values[$Name]);
                    }
                    else
                    {
                        $IsEmpty = !strlen(trim($Values[$Name]));
                    }
                    break;
            }

            # if field has validation function
            if (isset($Params["ValidateFunction"]))
            {
                # swap in our object if this is one of our methods
                $VFunc = $Params["ValidateFunction"];
                if (is_array($VFunc) && ($VFunc[0] == "FormUI"))
                {
                    $VFunc[0] = $this;
                }

                # call validation function for value
                $Args = array_merge(array($Name, $Values[$Name], $Values),
                        $this->ExtraValidationParams);
                $ErrMsg = call_user_func_array($VFunc, $Args);
                if ($ErrMsg === FALSE)
                {
                    throw new Exception("Calling validation function for"
                            ." parameter \"".$Name."\" failed.");
                }

                # log any resulting error
                if ($ErrMsg !== NULL)
                {
                    self::LogError($ErrMsg, $Name);
                    $ErrorsFound++;
                }
            }

            # if field is required and empty
            if ($IsEmpty && isset($Params["Required"]) && $Params["Required"])
            {
                # log error to indicate required value is missing
                self::LogError("<i>".$Params["Label"]."</i> is required.", $Name);
                $ErrorsFound++;
            }
            # else validate based on field type
            else
            {
                switch ($Params["Type"])
                {
                    case self::FTYPE_NUMBER:
                        # make sure value is within any specified range
                        if ((isset($Params["MinVal"])
                                        && ($Values[$Name] < $Params["MinVal"]))
                                || (isset($Params["MaxVal"])
                                        && ($Values[$Name] > $Params["MaxVal"])))
                        {
                            if (!isset($Params["MaxVal"]))
                            {
                                self::LogError("<i>".$Params["Label"]."</i> must be "
                                        .$Params["MinVal"]." or greater.", $Name);
                            }
                            elseif (!isset($Params["MinVal"]))
                            {
                                self::LogError("<i>".$Params["Label"]."</i> must be "
                                        .$Params["MaxVal"]." or less.", $Name);
                            }
                            else
                            {
                                self::LogError("<i>".$Params["Label"]."</i> must be"
                                        ." in the range ".$Params["MinVal"]
                                        ." to ".$Params["MaxVal"].".", $Name);
                            }
                            $ErrorsFound++;
                        }
                        break;

                    case self::FTYPE_URL:
                        # make sure URL entered looks valid
                        if (!$IsEmpty && (filter_var(
                                $Values[$Name], FILTER_VALIDATE_URL) === FALSE))
                        {
                            self::LogError("Value \"".$Values[$Name]
                                    ."\" does not appear to be a valid URL for <i>"
                                    .$Params["Label"]."</i>.", $Name);
                            $ErrorsFound++;
                        }
                        break;

                    case self::FTYPE_USER:
                        # make sure user name entered is valid
                        $UFactory = new CWUserFactory();
                        foreach ($Values[$Name] as $UId)
                        {
                            if (strlen($UId) && !$UFactory->UserExists($UId))
                            {
                                self::LogError("User ID \"".$UId
                                               ."\" not found for <i>"
                                               .$Params["Label"]."</i>.", $Name);
                                $ErrorsFound++;
                            }
                        }
                        break;
                }
            }
        }

        # report number of fields with invalid values found to caller
        return $ErrorsFound;
    }

    /**
    * Add values to be passed to input validation functions, in addition
    * to field name and value.
    * @see FormUI_Base::ValidateFieldInput()
    */
    public function AddValidationParameters()
    {
        $this->ExtraValidationParams = func_get_args();
    }

    /**
    * Retrieve values set by form.
    * @return array Array of configuration settings, with setting names
    *       for the index, and new setting values for the values.
    */
    public function GetNewValuesFromForm()
    {
        # for each form field
        $NewSettings = array();
        foreach ($this->FieldParams as $Name => $Params)
        {
            # skip header fields
            if ($Params["Type"] == self::FTYPE_HEADING) {  continue;  }

            # determine form field name (matches mechanism in HTML)
            $FieldName = $this->GetFormFieldName($Name,
                    ($Params["Type"] != self::FTYPE_PRIVILEGES));

            # assume the value will not change
            $DidValueChange = FALSE;
            $OldValue = isset($this->FieldValues[$Name])
                    ? $this->FieldValues[$Name]
                    : (isset($Params["Value"]) ? $Params["Value"] : NULL);
            $NewSettings[$Name] = $OldValue;

            # retrieve value based on configuration parameter type
            switch ($Params["Type"])
            {
                case self::FTYPE_FLAG:
                    # if radio buttons were used
                    if (array_key_exists("OnLabel", $Params)
                            && array_key_exists("OffLabel", $Params))
                    {
                        if (isset($_POST[$FieldName]))
                        {
                            $NewValue = ($_POST[$FieldName] == "1") ? TRUE : FALSE;

                            # flag that the values changed if they did
                            $DidValueChange = self::DidValueChange(
                                    $OldValue, $NewValue);

                            $NewSettings[$Name] = $NewValue;
                        }
                    }
                    # else checkbox was used
                    else
                    {
                        $NewValue = isset($_POST[$FieldName]) ? TRUE : FALSE;

                        # flag that the values changed if they did
                        $DidValueChange = self::DidValueChange($OldValue, $NewValue);

                        $NewSettings[$Name] = $NewValue;
                    }
                    break;

                case self::FTYPE_OPTION:
                    $NewValue = GetArrayValue($_POST, $FieldName, array());

                    # flag that the values changed if they did
                    $DidValueChange = self::DidValueChange($OldValue, $NewValue);

                    $NewSettings[$Name] = $NewValue;
                    break;

                case self::FTYPE_METADATAFIELD:
                    $NewValue = GetArrayValue($_POST, $FieldName, array());
                    if ($NewValue == "-1") {  $NewValue = array();  }

                    # flag that the values changed if they did
                    $DidValueChange = self::DidValueChange($OldValue, $NewValue);

                    $NewSettings[$Name] = $NewValue;
                    break;

                case self::FTYPE_PRIVILEGES:
                    $Schemas = GetArrayValue($Params, "Schemas");
                    $MFields = GetArrayValue($Params, "MetadataFields", array());
                    $PEditor = new PrivilegeEditingUI($Schemas, $MFields);
                    $NewValues = $PEditor->GetPrivilegeSetsFromForm();
                    $NewValue = $NewValues[$FieldName];
                    $DidValueChange = self::DidValueChange($OldValue, $NewValue);
                    $NewSettings[$Name] = $NewValue;
                    break;

                case self::FTYPE_SEARCHPARAMS:
                    $SPEditor = new SearchParameterSetEditingUI($FieldName);
                    $NewValue = $SPEditor->GetValuesFromFormData();
                    $DidValueChange = self::DidValueChange($OldValue, $NewValue);
                    $NewSettings[$Name] = $NewValue;
                    break;

                case self::FTYPE_FILE:
                    $NewSettings[$Name] = GetArrayValue(
                            $_POST, $FieldName."_ID", array());
                    foreach ($NewSettings[$Name] as $Index => $FileId)
                    {
                        if ($FileId == self::NO_VALUE_FOR_FIELD ||
                            in_array($FileId, $this->DeletedFiles))
                        {
                            unset($NewSettings[$Name][$Index]);
                        }
                    }

                    # if we have any files in ExtraValues after a HandleUploads,
                    # include those too
                    $ExtraSettings = GetArrayValue(
                        $this->ExtraValues, $FieldName."_ID", []);
                    foreach ($ExtraSettings as $FileId)
                    {
                        if (!in_array($FileId, $NewSettings[$Name]) &&
                            !in_array($FileId, $this->DeletedFiles))
                        {
                            $NewSettings[$Name][]= $FileId;
                        }
                    }

                    break;

                case self::FTYPE_IMAGE:
                    $NewSettings[$Name] = GetArrayValue(
                            $_POST, $FieldName."_ID", array());

                    foreach ($NewSettings[$Name] as $Index => $ImageId)
                    {
                        if ($ImageId == self::NO_VALUE_FOR_FIELD ||
                            in_array($ImageId, $this->DeletedImages) )
                        {
                            unset($NewSettings[$Name][$Index]);
                        }
                        elseif (isset($_POST[$FieldName."_AltText_".$ImageId]))
                        {
                            $Image = new SPTImage($ImageId);
                            $Image->AltText($_POST[$FieldName."_AltText_".$ImageId]);
                        }
                    }

                    # if we have any images in ExtraValues after a HandleUploads,
                    # include those too
                    $ExtraSettings = GetArrayValue(
                        $this->ExtraValues, $FieldName."_ID", []);
                    foreach ($ExtraSettings as $ImageId)
                    {
                        if (!in_array($ImageId, $NewSettings[$Name]) &&
                            !in_array($ImageId, $this->DeletedImages))
                        {
                            $NewSettings[$Name][]= $ImageId;
                            $AltTextKey = $FieldName."_AltText_".$ImageId;
                            if (isset($this->ExtraValues[$AltTextKey]))
                            {
                                $Image = new SPTImage($ImageId);
                                $Image->AltText(
                                    $this->ExtraValues[$AltTextKey]);
                            }
                        }
                    }

                    break;

                case self::FTYPE_USER:
                    if (isset($_POST[$FieldName]))
                    {
                        # filter blank entries out of input
                        # (these come up when a user was deleted from the list)
                        $NewValue = array_filter(
                            $_POST[$FieldName],
                            function($v) {
                                return strlen($v) > 0 ? TRUE : FALSE;
                            });

                        $DidValueChange = self::DidValueChange($OldValue, $NewValue);

                        $NewSettings[$Name] = $NewValue;
                    }
                    break;

                default:
                    if (isset($_POST[$FieldName]))
                    {
                        $NewValue = $_POST[$FieldName];

                        # flag that the values changed if they did
                        $DidValueChange = self::DidValueChange($OldValue, $NewValue);

                        $NewSettings[$Name] = $NewValue;
                    }
                    break;
            }

            # if value changed and there is an event to signal for changes
            if ($DidValueChange && $this->SettingChangeEventName)
            {
                # set info about changed value in event parameters if appropriate
                $EventParams = $this->SettingChangeEventParams;
                foreach ($EventParams as $ParamName => $ParamValue)
                {
                    switch ($ParamName)
                    {
                        case "SettingName":
                            $EventParams[$ParamName] = $Name;
                            break;

                        case "OldValue":
                            $EventParams[$ParamName] = $OldValue;
                            break;

                        case "NewValue":
                            $EventParams[$ParamName] = $NewValue;
                            break;
                    }
                }

                # signal event
                $GLOBALS["AF"]->SignalEvent(
                        $this->SettingChangeEventName, $EventParams);
            }
        }

        # return updated setting values to caller
        return $NewSettings;
    }

    /**
    * Get value for form field.
    * @param string $FieldName Canonical field name.
    * @return mixed Value or array of values for field.
    */
    public function GetFieldValue($FieldName)
    {
        # get base form field name
        $FieldType = $this->FieldParams[$FieldName]["Type"];
        $FormFieldName = $this->GetFormFieldName($FieldName,
                ($FieldType != self::FTYPE_PRIVILEGES));

        # get fallback value for field (in case no value from form)
        if (isset($this->FieldValues[$FieldName]))
        {
            $Value = $this->FieldValues[$FieldName];
        }
        elseif (isset($this->FieldParams[$FieldName]["Value"]))
        {
            $Value = self::LoadValue($FieldType,
                    $this->FieldParams[$FieldName]["Value"]);
        }
        elseif (isset($this->FieldParams[$FieldName]["Default"]))
        {
            $Value = self::LoadValue($FieldType,
                    $this->FieldParams[$FieldName]["Default"]);
        }
        else
        {
            $Value = self::LoadValue($FieldType, NULL);
        }

        switch ($FieldType)
        {
            case self::FTYPE_FILE:
            case self::FTYPE_IMAGE:
                # get name of image ID form field
                $FileIdFieldName = $FormFieldName."_ID";

                # use an updated value for this field if available
                if (isset($this->HiddenFields[$FileIdFieldName]))
                {
                    $Value = $this->HiddenFields[$FileIdFieldName];
                }
                # else use a value from form if available
                elseif (isset($_POST[$FileIdFieldName]))
                {
                    $Value = $_POST[$FileIdFieldName];
                }

                # add in any previously-set extra values
                if (isset($this->ExtraValues[$FileIdFieldName])
                        && count($this->ExtraValues[$FileIdFieldName]))
                {
                    if (!is_array($Value))
                    {
                        $Value = array($Value);
                    }
                    $Value = array_merge($Value,
                            $this->ExtraValues[$FileIdFieldName]);
                }
                break;

            case self::FTYPE_SEARCHPARAMS:
                # use incoming form value if available
                if (isset($_POST[$FormFieldName]))
                {
                    # use incoming form value
                    $SPEditor = new SearchParameterSetEditingUI($FormFieldName);
                    $Value = $SPEditor->GetValuesFromFormData();
                }
                break;

            case self::FTYPE_PRIVILEGES:
                # use incoming form value if available
                $Schemas = GetArrayValue(
                        $this->FieldParams[$FieldName], "Schemas");
                $MFields = GetArrayValue(
                        $this->FieldParams[$FieldName], "MetadataFields", array());
                $PEditor = new PrivilegeEditingUI($Schemas, $MFields);
                $PSet = $PEditor->GetPrivilegeSetFromForm($FormFieldName);
                if ($PSet instanceof PrivilegeSet)
                {
                    # use incoming form value
                    $Value = $PSet;
                }
                break;

            case self::FTYPE_QUICKSEARCH:
                $MField = new MetadataField(
                    $this->FieldParams[$FieldName]["Field"]);

                # use incoming values if available
                if (isset($_POST[$FormFieldName]))
                {
                    $Value = $this->ConvertItemIdsToNames(
                        $MField, $_POST[$FormFieldName]);
                }
                break;

            default:
                # use incoming form value if available
                if (isset($_POST[$FormFieldName]))
                {
                    # use incoming form value
                    $Value = $_POST[$FormFieldName];
                }
                break;
        }

        # return value found to caller
        return $Value;
    }

    /**
    * Handle image and file uploads.
    */
    public function HandleUploads()
    {
        # for each form field
        foreach ($this->FieldParams as $FieldName => $FieldParams)
        {
            # move on to next field if this field does not allow uploads
            if (($FieldParams["Type"] != self::FTYPE_FILE)
                    && ($FieldParams["Type"] != self::FTYPE_IMAGE))
            {
                continue;
            }

            # move on to next field if this field does not have an uploaded file
            $FormFieldName = $this->GetFormFieldName($FieldName);
            if (!isset($_FILES[$FormFieldName]["name"]))
            {
                continue;
            }
            $UploadedFileName = $_FILES[$FormFieldName]["name"];
            if (!strlen($UploadedFileName))
            {
                continue;
            }

            # create temp copy of file with correct name
            $TmpFile = "tmp/".$UploadedFileName;
            $CopyResult = copy($_FILES[$FormFieldName]["tmp_name"], $TmpFile);

            switch ($FieldParams["Type"])
            {
                case self::FTYPE_FILE:
                    # create new file object from uploaded file
                    $File = File::Create($TmpFile, $UploadedFileName);

                    # check for errors during file object creation
                    if (!is_object($File))
                    {
                        switch ($File)
                        {
                            case File::FILESTAT_ZEROLENGTH:
                                $this->LogError("Uploaded file "
                                        .$UploadedFileName." was empty"
                                        ." (zero length).", $FieldName);
                                break;

                            default:
                                $this->LogError("Error encountered with"
                                        ." uploaded file "
                                        .$UploadedFileName." ("
                                        .$File.").", $FieldName);
                                break;
                        }
                        unlink($TmpFile);
                    }
                    else
                    {
                        # add file ID to extra values
                        $this->ExtraValues[$FormFieldName."_ID"][] = $File->Id();
                    }
                    break;

                case self::FTYPE_IMAGE:
                    # create new image object from uploaded file
                    $Image = new SPTImage($TmpFile,
                            $FieldParams["MaxWidth"],
                            $FieldParams["MaxHeight"],
                            $FieldParams["MaxPreviewWidth"],
                            $FieldParams["MaxPreviewHeight"],
                            $FieldParams["MaxThumbnailWidth"],
                            $FieldParams["MaxThumbnailHeight"]);

                    # check for errors during image object creation
                    if ($Image->Status() != AI_OKAY)
                    {
                        switch ($Image->Status())
                        {
                            case AI_UNKNOWNTYPE:
                                $this->LogError("Unknown format for image file "
                                        .$UploadedFileName.".", $FieldName);
                                break;

                            case AI_UNSUPPORTEDFORMAT:
                                $this->LogError("Unsupported format for image file "
                                        .$UploadedFileName." ("
                                        .$Image->Format().").", $FieldName);
                                break;

                            default:
                                $this->LogError("Error encountered when"
                                        ." processing uploaded image file "
                                        .$UploadedFileName.".", $FieldName);
                                break;
                        }
                        unlink($TmpFile);
                    }
                    else
                    {
                        # set image object alternate text
                        $Image->AltText($_POST[$FormFieldName."_AltText_NEW"]);

                        # add image ID to extra values
                        $this->ExtraValues[$FormFieldName."_ID"][] = $Image->Id();
                    }
                    break;
            }
        }
    }

    /**
    * Handle image and file deletions.
    */
    public function HandleDeletes()
    {
        # if image ID to delete was supplied
        $DeleteFieldName = $this->GetFormFieldName("ImageToDelete");
        $IncomingValue = GetFormValue($DeleteFieldName);
        if (is_numeric($IncomingValue)
                || (is_array($IncomingValue) && count($IncomingValue)))
        {
            # retrieve ID of image
            $ImageId = is_array($IncomingValue)
                    ? array_shift($IncomingValue)
                    : $IncomingValue;

            # add ID to deleted images list
            if (is_numeric($ImageId))
            {
                $this->DeletedImages[] = $ImageId;
            }
        }

        # if file ID to delete was supplied
        $DeleteFieldName = $this->GetFormFieldName("FileToDelete");
        $IncomingValue = GetFormValue($DeleteFieldName);
        if (is_numeric($IncomingValue)
                || (is_array($IncomingValue) && count($IncomingValue)))
        {
            # retrieve ID of file
            $FileId = is_array($IncomingValue)
                    ? array_shift($IncomingValue)
                    : $IncomingValue;

            # add ID to deleted files list
            if (is_numeric($FileId))
            {
                $this->DeletedFiles[] = $FileId;
            }
        }
    }

    /**
    * Set event to signal when retrieving values from form when settings
    * have changed.  If the supplied event parameters include parameter
    * names (indexes) of "SettingName", "OldValue", or "NewValue", the
    * parameter value will be replaced with an appropriate value before
    * the event is signaled.
    * @param string $EventName Name of event to signal.
    * @param array $EventParams Array of event parameters, with CamelCase
    *       parameter names for index.  (OPTIONAL)
    * @see FormUI_Base::GetNewsettingsFromForm()
    */
    public function SetEventToSignalOnChange($EventName, $EventParams = array())
    {
        $this->SettingChangeEventName = $EventName;
        $this->SettingChangeEventParams = $EventParams;
    }

    /**
    * Determine if a new form field value is different from an old one.
    * @param mixed $OldValue Old field value.
    * @param mixed $NewValue New field value.
    * @return Returns TRUE if the values are different and FALSE otherwise.
    */
    public static function DidValueChange($OldValue, $NewValue)
    {
        # didn't change if they are identical
        if ($OldValue === $NewValue)
        {
            return FALSE;
        }

        # need special cases from this point because PHP returns some odd results
        # when performing loose equality comparisons:
        # http://php.net/manual/en/types.comparisons.php#types.comparisions-loose

        # consider NULL and an empty string to be the same. this is in case a field
        # is currently set to NULL and receives an empty value from the form.
        # $_POST values are always strings
        if ((is_null($OldValue) && is_string($NewValue) && !strlen($NewValue))
            || (is_null($NewValue) && is_string($OldValue) && !strlen($OldValue)))
        {
            return FALSE;
        }

        # if they both appear to be numbers and are equal
        if (is_numeric($OldValue) && is_numeric($NewValue) && $OldValue == $NewValue)
        {
            return FALSE;
        }

        # true-like values
        if (($OldValue === TRUE && ($NewValue === 1 || $NewValue === "1"))
            || ($NewValue === TRUE && ($OldValue === 1 || $OldValue === "1")))
        {
            return FALSE;
        }

        # false-like values
        if (($OldValue === FALSE && ($NewValue === 0 || $NewValue === "0"))
            || ($NewValue === FALSE && ($OldValue === 0 || $OldValue === "0")))
        {
            return FALSE;
        }

        # arrays
        if (is_array($OldValue) && is_array($NewValue))
        {
            # they certainly changed if the counts are different
            if (count($OldValue) != count($NewValue))
            {
                return TRUE;
            }

            # the algorithm for associative arrays is slightly different from
            # sequential ones. the values for associative arrays must match the keys
            if (count(array_filter(array_keys($OldValue), "is_string")))
            {
                foreach ($OldValue as $Key => $Value)
                {
                    # it changed if the keys don't match
                    if (!array_key_exists($Key, $NewValue))
                    {
                        return TRUE;
                    }

                    # the arrays changed if a value changed
                    if (self::DidValueChange($Value, $NewValue[$Key]))
                    {
                        return TRUE;
                    }
                }
            }

            # sequential values don't have to have the same keys, just the same
            # values
            else
            {
                # sort them so all the values match up if they're equal
                sort($OldValue);
                sort($NewValue);

                foreach ($OldValue as $Key => $Value)
                {
                    # the arrays changed if a value changed
                    if (self::DidValueChange($Value, $NewValue[$Key]))
                    {
                        return TRUE;
                    }
                }
            }

            # the arrays are equal
            return FALSE;
        }

        # they changed
        return TRUE;
    }

    /**
    * Load value of requested type from supplied data.
    * @param string $Type Type of value (FTYPE_*).
    * @param mixed $Data Data to use in loading value.
    * @return mixed Loaded value.
    */
    static public function LoadValue($Type, $Data)
    {
        switch ($Type)
        {
            case self::FTYPE_FILE:
            case self::FTYPE_IMAGE:
                $Value = ($Data === NULL) ? array() : $Data;
                break;

            case self::FTYPE_PRIVILEGES:
                $Value = ($Data instanceof PrivilegeSet) ? $Data
                        : new PrivilegeSet($Data);
                break;

            case self::FTYPE_SEARCHPARAMS:
                $Value = ($Data instanceof SearchParameterSet) ? $Data
                        : new SearchParameterSet($Data);
                break;

            default:
                $Value = $Data;
                break;
        }

        return $Value;
    }

    /**
    * Validate value as valid-appearing email address.  This is intended to be
    * used with the "ValidateFunction" capability, like this:
    * @code
    *     "ValidateFunction" => array("FormUI", "ValidateEmail"),
    * @endcode
    * (The "FormUI" part will be replaced by the appropropriate FormUI object
    * before the method is called.)
    * @param string $FieldName Name of form field.
    * @param array $FieldValues Form values being validated.
    * @return string Error message or NULL if value appears valid.
    */
    public function ValidateEmail($FieldName, $FieldValues)
    {
        if (!is_array($FieldValues)) {  $FieldValues = array($FieldValues);  }
        foreach ($FieldValues as $Value)
        {
            if (trim($Value) == "") {  continue;  }
            if (filter_var($Value, FILTER_VALIDATE_EMAIL) === FALSE)
            {
                return "The value for <i>".$this->FieldParams[$FieldName]["Label"]
                        ."</i> does not appear to be a valid email address.";
            }
        }
        return NULL;
    }

    /**
    * Validate value as valid-appearing URL.  This is intended to be used
    * with the "ValidateFunction" capability, like this:
    * @code
    *     "ValidateFunction" => array("FormUI", "ValidateUrl"),
    * @endcode
    * (The "FormUI" part will be replaced by the appropropriate FormUI object
    * before the method is called.)
    * @param string $FieldName Name of form field.
    * @param array $FieldValues Form values being validated.
    * @return string Error message or NULL if value appears valid.
    */
    public function ValidateUrl($FieldName, $FieldValues)
    {
        if (!is_array($FieldValues)) {  $FieldValues = array($FieldValues);  }
        foreach ($FieldValues as $Value)
        {
            if (trim($Value) == "") {  continue;  }
            if (filter_var($Value, FILTER_VALIDATE_URL) === FALSE)
            {
                return "The value for <i>".$this->FieldParams[$FieldName]["Label"]
                        ."</i> does not appear to be a valid URL.";
            }
        }
        return NULL;
    }

    /**
    * Validate value as valid host name (i.e. one that can be resolved to an
    * IP address via DNS).  This is intended to be used with the "ValidateFunction"
    * capability, like this:
    * @code
    *     "ValidateFunction" => array("FormUI", "ValidateHostName"),
    * @endcode
    * (The "FormUI" part will be replaced by the appropropriate FormUI object
    * before the method is called.)
    * @param string $FieldName Name of form field.
    * @param array $FieldValues Form values being validated.
    * @return string Error message or NULL if value appears valid.
    */
    public function ValidateHostName($FieldName, $FieldValues)
    {
        if (!is_array($FieldValues)) {  $FieldValues = array($FieldValues);  }
        foreach ($FieldValues as $Value)
        {
            if (trim($Value) == "") {  continue;  }
            if (gethostbyname($Value) === $Value)
            {
                return "The value for <i>".$this->FieldParams[$FieldName]["Label"]
                        ."</i> does not appear to be a valid host name.";
            }
        }
        return NULL;
    }


    # ---- PRIVATE INTERFACE -------------------------------------------------

    protected $DeletedFiles = array();
    protected $DeletedImages = array();
    protected $ExtraValidationParams = array();
    protected $ExtraValues = array();
    protected $FieldParams;
    protected $FieldValues;
    protected $HiddenFields = array();
    protected $SettingChangeEventName = NULL;
    protected $SettingChangeEventParams = array();

    protected static $ErrorMessages = array();

    /** Marker used to indicate currently no value for field. */
    const NO_VALUE_FOR_FIELD = "NO VALUE";

    /**
    * Display HTML form field for specified field.
    * @param string $Name Field name.
    * @param mixed $Value Current value for field.
    * @param array $Params Field parameters.
    */
    abstract protected function DisplayFormField($Name, $Value, $Params);

    /**
    * Get HTML form field name for specified field.
    * @param string $FieldName Field name.
    * @param bool $IncludePrefix If TRUE, "F_" prefix is included.  (OPTIONAL,
    *       defaults to TRUE.)
    * @return string Form field name.
    */
    protected function GetFormFieldName($FieldName, $IncludePrefix = TRUE)
    {
        return ($IncludePrefix ? "F_" : "")
                .($this->UniqueKey ? $this->UniqueKey."_" : "")
                .preg_replace("/[^a-zA-Z0-9]/", "", $FieldName);
    }

    /**
    * Get HTML for hidden form fields associated with form processing.
    */
    protected function GetHiddenFieldsHtml()
    {
        $Html = "";
        if (count($this->HiddenFields))
        {
            foreach ($this->HiddenFields as $FieldName => $Value)
            {
                if (is_array($Value))
                {
                    foreach ($Value as $EachValue)
                    {
                        $Html .= '<input type="hidden" name="'.$FieldName
                                .'[]" value="'.htmlspecialchars($EachValue).'">';
                    }
                }
                else
                {
                    $Html .= '<input type="hidden" name="'.$FieldName
                        .'[]" id="'.$FieldName.'" value="'
                        .htmlspecialchars($Value).'">';
                }
            }
        }
        return $Html;
    }

    /**
    * Take an array of ItemIds and convert it to [ ItemId => ItemName ].
    * @param MetadataField $MField Field giving the namespace of the ItemIds.
    * @param array $ItemIds Int item ids.
    * @return array keyed by ItemId with values giving item names.
    */
    protected function ConvertItemIdsToNames($MField, $ItemIds)
    {
        # define $ExistsFn to determine if an ItemId exists and
        # $NameFn to get its name
        switch ($MField->Type())
        {
            case MetadataSchema::MDFTYPE_TREE:
            case MetadataSchema::MDFTYPE_CONTROLLEDNAME:
                $Factory = $MField->GetFactory();

                $ExistsFn = function($Id) use ($Factory) {
                    return $Factory->ItemExists($Id);
                };

                $NameFn = function ($Id) use ($Factory) {
                    return ($Factory->GetItem($Id)->Name());
                };
                break;

            case MetadataSchema::MDFTYPE_USER:
                $Factory = new CWUserFactory();

                $ExistsFn = function($Id) use ($Factory) {
                    return $Factory->UserExists($Id);
                };

                $NameFn = function ($Id) {
                    return (new CWUser($Id))->Name();
                };
                break;

            case MetadataSchema::MDFTYPE_REFERENCE:
                $ExistsFn = function($Id) {
                    return Resource::ItemExists($Id);
                };

                $NameFn = function ($Id) {
                    return (new Resource($Id))->GetMapped("Title");
                };
                break;

            default:
                throw new Exception(
                    "Invalid field type for ItemId to ItemName conversion.");
        }

        # iterate over incoming values, extracting names for all the ones that exist
        $Value = [];
        foreach ($ItemIds as $Id)
        {
            if ($ExistsFn($Id))
            {
                $Value[$Id] = $NameFn($Id);
            }
        }

        return $Value;
    }
}
