<?PHP
#
#   FILE:  CWUser.php
#
#   Part of the Collection Workflow Integration System (CWIS)
#   Copyright 2013 Edward Almasy and Internet Scout Research Group
#   http://scout.wisc.edu/cwis/
#

/**
* CWIS-specific user class.
*/
class CWUser extends User
{

    # ---- PUBLIC INTERFACE --------------------------------------------------

    /**
    * Load user data from the given user info or from the session if available.
    * @param mixed $UserInfo A user ID or user name. (OPTIONAL)
    */
    public function __construct($UserInfo=NULL)
    {
        static $EmailWrapperSet = FALSE;
        if (!$EmailWrapperSet)
        {
            User::SetEmailFunction(array("CWUser", "EmailWrapper"));
            $EmailWrapperSet = TRUE;
        }

        parent::__construct($UserInfo);

        if ($this->Result !== U_OKAY)
        {
            throw new Exception(
                "Unable to load user information.");
        }

        # try to fetch the associated resource if the user was found
        if (!$this->IsAnonymous())
        {
            $Resource = $this->FetchAssociatedResource($this->UserId);

            # the associated resource was successfully found
            if ($Resource instanceof Resource)
            {
                $this->Resource = $Resource;
            }
            # there was a problem finding the resource
            else
            {
                throw new Exception(
                    "Unable to load corresponding resource for user.");
            }
        }
        else
        {
            $this->Resource = NULL;
        }
    }

    /**
    * Log the specified user in and associate the underlying Resource
    *     with this CWUser.
    * @param string $UserName User to log in.
    * @param string $Password User's password.
    * @param bool $IgnorePassword TRUE to skip password validation
    *     (OPTIONAL, default FALSE)
    */
    public function Login($UserName, $Password, $IgnorePassword = FALSE)
    {
        parent::Login($UserName, $Password, $IgnorePassword);

        if ($this->Result == U_OKAY)
        {
            $Resource = $this->FetchAssociatedResource($this->UserId);

            # the associated resource was successfully found
            if ($Resource instanceof Resource)
            {
                $this->Resource = $Resource;
            }
            # there was a problem finding the resource
            else
            {
                throw new Exception(
                    "Unable to load corresponding resource for user.");
            }
        }

        return $this->Result;
    }

    /**
    * Log this user out and disassociate their underlying Resource from this CWUser.
    */
    public function Logout()
    {
        parent::Logout();
        $this->Resource = NULL;
    }

    /**
    * THIS FUNCTION HAS BEEN DEPRECATED
    * This provides compatibility for interfaces written to use a
    * version of PrivilegeSet from CWIS 3.0.0 to 3.1.0.
    * @param PrivilegeSet $NewValue New value (OPTIONAL, default NULL)
    * @return PrivilegeSetCompatibiliityShim for use in legacy code.
    */
    public function Privileges(PrivilegeSet $NewValue = NULL)
    {
        if ($NewValue !== NULL)
        {
            throw new Exception(
                "Attempt to set user privileges with CWUser::Privileges(), "
                ."which is no longer supported");
        }

        return new PrivilegeSetCompatibilityShim($this);
    }

    /**
    * Get the ID of the user resource associated with the user.
    * @return Returns the ID of the associated user resource or NULL if it's not
    *      available, e.g., the user isn't logged in.
    */
    public function ResourceId()
    {
        return ($this->Resource !== NULL) ? $this->Resource->Id() : NULL;
    }

    /**
    * Get the associated user resource for this user.
    * @return Returns the associated user resource or NULL if it's not
    *      available, e.g., the user isn't logged in.
    */
    public function GetResource()
    {
        return ($this->Resource !== NULL) ? $this->Resource : NULL;
    }

    /**
    * Determine if a user has a given privilege, or satisfies the
    * conditions specified by a given privilege set.  Calling this
    * function with a PrivilegeSet as the first argument is supported
    * only for backwards compatibility.  New code should not do this.
    * @param mixed $Privilege Privilige number or Privilige set to check
    * @param mixed $Privileges Additional privileges (as in
    *      parent::HasPriv()), or a Resource to use if the first arg was a
    *      PrivilegeSet.
    * @return TRUE if the user has the specified privilege (or satisfies
    *      the requirements of the specified privilege set.
    */
    public function HasPriv($Privilege, $Privileges = NULL)
    {
        if ($Privilege instanceof PrivilegeSet)
        {
            if ($Privileges instanceof Resource)
            {
                return $Privilege->MeetsRequirements($this, $Privileges);
            }
            else
            {
                return $Privilege->MeetsRequirements($this);
            }
        }
        else
        {
            $Args = func_get_args();
            if (in_array(PRIV_ISLOGGEDIN, $Args) && $this->IsLoggedIn())
            {
                return TRUE;
            }

            $Args = self::FilterPrivileges($Args);
            if (count($Args)>0)
            {
                return call_user_func_array( "parent::HasPriv", $Args);
            }

            return FALSE;
        }
    }

    /**
    * Grant privilege to a a user.
    * @param int $Privilege Privilege to grant.
    * @throws Exception On attempts to grant a pseudo-privilege.
    * @see User::GrantPriv() for return values.
    */
    public function GrantPriv($Privilege)
    {
        if (self::IsPseudoPrivilege($Privilege))
        {
            throw new Exception("Attempt to grant pseudo-privilege to user");
        }

        return parent::GrantPriv($Privilege);
    }

    /**
    * Clear current user privs and replace them with the specified list.
    * @param array $NewPrivileges New privilege list to
    *   assign. Pseudo-privileges are filtered from the list passed to
    *   User::SetPrivList().
    */
    public function SetPrivList($NewPrivileges)
    {
        parent::SetPrivList( self::FilterPrivileges($NewPrivileges) );
    }

    /**
    * Adapter method to bridge between User class and Email class.
    * @param string $To To line for message.
    * @param string $Subject Subject line for message.
    * @param string $Message Body of message.
    * @param string $AdditionalHeaders Other message header lines, concatenated
    *       together into a string.
    * @return bool TRUE if message successfully accepted for delivery,
    *       otherwise FALSE.
    */
    public static function EmailWrapper($To, $Subject, $Message, $AdditionalHeaders)
    {
        # extract "From" address from supplied headers if available
        if (strlen($AdditionalHeaders))
        {
            $HeaderLines = explode("\n", $AdditionalHeaders);
            $Headers = array();
            foreach ($HeaderLines as $Line)
            {
                $HeaderLine = trim($Line);
                if (preg_match("/^from:/i", $Line))
                {
                    $From = preg_replace("/^from:/i", "", $Line);
                }
                else
                {
                    $Headers[] = $HeaderLine;
                }
            }
        }

        # send message
        $Msg = new Email();
        if (isset($From)) {  $Msg->From($From);  }
        $Msg->To($To);
        $Msg->Subject($Subject);
        $Msg->AddHeaders($Headers);
        $Msg->Body($Message);
        $Result = $Msg->Send();

        # report success to caller
        return $Result;
    }

    /**
    * Get all custom user fields.
    * @return Returns an array of the custom user fields.
    */
    public static function GetCustomUserFields()
    {
        static $CustomFields;

        if (!isset($CustomFields))
        {
            $CustomFields = array();
            $Schema = new MetadataSchema(MetadataSchema::SCHEMAID_USER);

            foreach ($Schema->GetFields() as $Field)
            {
                # they're custom if not owned by CWIS
                if ($Field->Owner() != "CWISCore")
                {
                    $CustomFields[$Field->Id()] = $Field;
                }
            }
        }

        return $CustomFields;
    }

    /**
    * Get the default user fields.
    * @return Returns an array of the default user fields.
    */
    public static function GetDefaultUserFields()
    {
        static $DefaultFields;

        if (!isset($DefaultFields))
        {
            $DefaultFields = array();
            $Schema = new MetadataSchema(MetadataSchema::SCHEMAID_USER);

            foreach ($Schema->GetFields() as $Field)
            {
                # they're default if owned by CWIS
                if ($Field->Owner() == "CWISCore")
                {
                    $DefaultFields[$Field->Id()] = $Field;
                }
            }
        }

        return $DefaultFields;
    }

    # ---- OVERRIDDEN METHODS ------------------------------------------------

    /**
    * Delete the user and its associated user resource. Methods should not be
    * called on the object after calling this method.
    * @return Returns the status of the deletion attempt.
    */
    public function Delete()
    {
        # delete the associated user resource if set
        if (isset($this->Resource))
        {
            $this->Resource->Delete();
            $this->Result = U_OKAY;
        }

        return parent::Delete();
    }

    /**
    * Get a value from the specified field.
    * @param string $FieldName The name of the field to get.
    * @return Returns the field value or NULL if user data isn't available,
    *      e.g., the user isn't logged in.
    */
    public function Get($FieldName)
    {
        # all values are NULL for anonymous users
        if ($this->IsAnonymous())
        {
            return NULL;
        }

        if (in_array($FieldName, self::$FieldsOnlyInAPUsers))
        {
            return parent::Get($FieldName);
        }
        else
        {
            return $this->Resource->Get($FieldName);
        }
    }

    /**
    * Set a value for the specified field.
    * @param string $Field The field to set (name, id, or MetadataField object).
    * @param mixed $NewValue The value to which to set the field.
    * @return Returns the status of the operation.
    */
    public function Set($Field, $NewValue)
    {
        if ($this->IsAnonymous())
        {
            throw new Exception(
                "Attempt to set User field value when "
                ."no user is logged in.");
        }

        # make sure Field is a FieldName
        if ($Field instanceof MetadataField)
        {
            $Field = $Field->Name();
        }

        # if this field is not among those that should only exists in
        # the APUsers table
        if (!in_array($Field, self::$FieldsOnlyInAPUsers))
        {
            # set it in our corresponding resource
            $this->Resource->Set($Field, $NewValue);
        }

        # if the given field exists in the APUsers table, update that too
        if ($this->DB->FieldExists("APUsers", $Field))
        {
            parent::Set($Field, $NewValue);
        }
        else
        {
            # indicate success for fields that don't have a column in APUsers
            $this->Result = U_OKAY;
        }

        return $this->Result;
    }

    # ---- PRIVATE INTERFACE -------------------------------------------------

    /**
    * The user resource associated with the user or NULL if the user isn't
    * logged in.
    */
    protected $Resource = NULL;

    # list of fields that exist in APUsers that are not mirrored as MetadataFields
    private static $FieldsOnlyInAPUsers = [
        # fields necessary to for user identification
        "UserId", "UserName", "EMail", "EMailNew",

        # fields necessary for authentication
        "UserPassword", "RegistrationConfirmed",

        # fields that can't be in a schema because they are updated by User
        "LastLoginDate", "LastActiveDate", "LastIPAddress", "LastLocation", "LoggedIn",

        # user preferences
        "ActiveUI", "BrowsingFieldId", "RecordsPerPage", "SearchSelections",
    ];

    /**
    * Fetch the associated user resource based off of a user ID.
    * @param int $UserId The user ID for the user associated with the esource.
    * @return Returns the associated user resource or an error status if an
    *      error occurs.
    */
    protected function FetchAssociatedResource($UserId)
    {
        if (self::$UserIdFieldId === NULL)
        {
            # get the user schema
            $Schema = new MetadataSchema(MetadataSchema::SCHEMAID_USER);

            # pull out the UserId field, which should only be one
            $Field = $Schema->GetField("UserId");

            # and get its FieldId
            self::$UserIdFieldId = intval($Field->Id());
        }

        # find the matching Resources (should only be one)
        $this->DB->Query(
            "SELECT ResourceId FROM ResourceUserInts WHERE ".
            "FieldId=".self::$UserIdFieldId.
            " AND UserId=".intval($UserId) );
        $ResourceIds = $this->DB->FetchColumn("ResourceId");
        $ResourceIdCount = count($ResourceIds);

        # no resource found
        if ($ResourceIdCount < 1)
        {
            return U_NOSUCHUSER;
        }

        # too many resources found
        if ($ResourceIdCount > 1)
        {
            throw new Exception(
                "Multiple resources exist for a single user, "
                ."which should be impossible");
        }

        # construct the associated resource and return it
        return new Resource(array_shift($ResourceIds));
    }

    /**
    * Determine if a privilege is one of the standard privileges built
    * in to CWIS.
    * @param int $Priv Privilege to check.
    * @return TRUE for standard privileges, FALSE otherwise
    */
    public static function IsStandardPrivilege($Priv)
    {
        return (self::MIN_STANDARD_PRIV <= $Priv &&
                $Priv <= self::MAX_STANDARD_PRIV) ? TRUE : FALSE;
    }

    /**
    * Determine if a privilege is one of the pseudo-privileges used by
    * CWIS.
    * @param int $Priv Privilege to check.
    * @return TRUE for pseudo-privileges, FALSE otherwise
    */
    public static function IsPseudoPrivilege($Priv)
    {
        return (self::MIN_PSEUDO_PRIV <= $Priv &&
                $Priv <= self::MAX_PSEUDO_PRIV) ? TRUE : FALSE;
    }

    /**
    * Determine if a privilege is a custom privilege created by a user
    * or by a plugin.
    * @param int $Priv Privilege to check.
    * @return TRUE for custom privileges, FALSE otherwise
    */
    public static function IsCustomPrivilege($Priv)
    {
        return (self::MIN_CUSTOM_PRIV <= $Priv &&
                $Priv <= self::MAX_CUSTOM_PRIV) ? TRUE : FALSE;
    }

    /**
    * Filter pseudo-privileges out of a privilege list.
    * @param array $Privs List to filter.
    * @return array Filtered list.
    */
    protected function FilterPrivileges($Privs)
    {
        $Result = [];
        foreach ($Privs as $Priv)
        {
            if (!self::IsPseudoPrivilege($Priv))
            {
                $Result[]= $Priv;
            }
        }

        return $Result;
    }

    const MIN_STANDARD_PRIV = 1;
    const MAX_STANDARD_PRIV = 74;

    const MIN_PSEUDO_PRIV = 75;
    const MAX_PSEUDO_PRIV = 99;

    const MIN_CUSTOM_PRIV = 100;
    const MAX_CUSTOM_PRIV = PHP_INT_MAX;

    # ---- MAINTAINED FOR BACKWARD COMPATIBILITY IN INTERFACES (BEGIN)

    # ---- user interface preference mnemonics
    # color avoidance flags
    const UIPREF_AVOID_RED =           1;
    const UIPREF_AVOID_REDGREEN =      2;
    const UIPREF_AVOID_BLUEYELLOW =    4;
    const UIPREF_AVOID_GREENYELLOW =   8;
    const UIPREF_AVOID_ORANGE =        16;
    const UIPREF_AVOID_REDBLACK =      32;
    const UIPREF_AVOID_PURPLEGREY =    64;
    const UIPREF_AVOID_USEMAXMONOCHR = 128;

    # content display options
    const UIPREF_CONTENTDENSITY_NOPREFERENCE =  0;
    const UIPREF_CONTENTDENSITY_DETAILED =  1;
    const UIPREF_CONTENTDENSITY_OVERVIEW =  2;

    # content view options
    const UIPREF_CONTENTVIEW_NOPREFERENCE =  0;
    const UIPREF_CONTENTVIEW_TEXTINTENSIVE =  1;
    const UIPREF_CONTENTVIEW_IMAGEINTENSIVE =  2;

    # audio description options
    const UIPREF_AUDIODESCRIPTION_NONE =  0;
    const UIPREF_AUDIODESCRIPTION_STANDARD =  1;
    const UIPREF_AUDIODESCRIPTION_EXPANDED =  2;

    # caption type options
    const UIPREF_CAPTIONTYPE_NONE =  0;
    const UIPREF_CAPTIONTYPE_VERBATIM =  1;
    const UIPREF_CAPTIONTYPE_REDUCEDREADINGLEVEL =  2;

    // @codingStandardsIgnoreStart

    # user interface / accessibility preferences
    function PrefFontSize($NewValue = DB_NOVALUE)
    {  return 0;  }

    function PrefFontTypeFace($NewValue = DB_NOVALUE)
    {  return 0;  }

    function PrefFontColor($NewValue = DB_NOVALUE)
    {  return 0;  }

    function PrefBackgroundColor($NewValue = DB_NOVALUE)
    {  return 0;  }

    function PrefColorAvoidanceFlags($NewValue = DB_NOVALUE)
    {  return 0;  }

    function PrefContentDensity($NewValue = DB_NOVALUE)
    {  return 0;  }

    function PrefContentView($NewValue = DB_NOVALUE)
    {  return 0;  }

    function PrefAudioDescriptionLevel($NewValue = DB_NOVALUE)
    {  return 0;  }

    function PrefAudioDescriptionLanguage($NewValue = DB_NOVALUE)
    { return 0;  }

    function PrefVisualDescriptionLanguage($NewValue = DB_NOVALUE)
    { return 0;  }

    function PrefImageDescriptionLanguage($NewValue = DB_NOVALUE)
    {  return 0;  }

    function PrefUseGraphicAlternatives($NewValue = DB_NOVALUE)
    {  return 0;  }

    function PrefSignLanguage($NewValue = DB_NOVALUE)
    {  return 0;  }

    function PrefCaptionType($NewValue = DB_NOVALUE)
    {  return 0;  }

    function PrefCaptionRate($NewValue = DB_NOVALUE)
    {  return 0;  }

    // @codingStandardsIgnoreEnd
    # ---- MAINTAINED FOR BACKWARD COMPATIBILITY IN INTERFACES (END)

    private static $UserIdFieldId = NULL;
}
